/*
 * $Copyright (c) 2014 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */
package com.apama.sample;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import com.apama.adapters.el.api.ELMappingExtensionManager;
import com.apama.adapters.el.api.ELMappingExtensionProvider;
import com.apama.event.Event;
import com.apama.event.parser.BooleanFieldType;
import com.apama.event.parser.DecimalFieldType;
import com.apama.event.parser.DictionaryFieldType;
import com.apama.event.parser.EventType;
import com.apama.event.parser.FieldType;
import com.apama.event.parser.FloatFieldType;
import com.apama.event.parser.IntegerFieldType;
import com.apama.event.parser.SequenceFieldType;
import com.apama.event.parser.StringFieldType;

/**
 * A sample Mapping Extension provider which implements ELMappingExtensionProvider and registers the convertToJson method used in mapping.
 */
public class ApamaJsonCreator implements ELMappingExtensionProvider {
	private static final String OPEN_CURLY = "{";
	private static final String CLOSE_CURLY = "}";
	private static final String OPEN_SQUARE = "[";
	private static final String CLOSE_SQUARE = "]";
	private static final String QUOTES = "\"";
	private static final String COMMA = ",";
	
	// register convertToJson method to mapping layer.
	@Override
	public void registerExtensions(ELMappingExtensionManager mappingExtensionManager) throws Exception {
		// register the convertToJson method as json:convertToJson. The method takes Apama event instance as parameter.
		mappingExtensionManager.registerMethod("json:convertToJson", getClass().getMethod("convertToJson", Event.class));
	}
	
	/**
	 * Converts the Apama event into JSON representation. Following rules are followed:
	 * <ul>
	 * 		<li>Only supported field types are primitive, sequence, event and dictionary. The key type of dictionary type field must be string.</li>
	 * 		<li>String type fields are represented as "fieldName":"fieldValue"</li>
	 * 		<li>Other primitive type fields are represented as "fieldName":fieldValue</li>
	 * 		<li>Sequence type fields are represented as "fieldName":[val1, val2], where representation of elements value depends on the type of field.</li>
	 * 		<li>Event type fields are represented as "fieldName":{"f1":v1, "f2":v2}</li>
	 * 		<li>Dictionary type fields are represented as "fieldName":{"key1":val1, "key2":val2}</li>
	 * </ul>
	 * @param event
	 * @return JSON string of Apama event.
	 */
	public static String convertToJson(Event event) {
		StringBuilder sb = new StringBuilder();
		processEventField(event, sb);
		return sb.toString();
	}
	
	// represent value of event type fields
	private static void processEventField(Event event, StringBuilder sb) {
		sb.append(OPEN_CURLY);
		if (event != null) {
			EventType eventType = event.getEventType();
			String[] fieldNames = eventType.getFieldNames();
			
			for (int i = 0; i < fieldNames.length; i++) {
				String fieldName = fieldNames[i];
				// Do not include destination field in JSON representation as it is used only for 
				// mapping to correct JMS:destination
				if (fieldName.equals("destination")) {
					continue;
				}
				
				Object value = event.getField(fieldName);
				FieldType<?> fieldType = eventType.getField(fieldName).getType();
				
				appendFieldname(fieldName, sb);
				handleFieldType(fieldType, value, sb);
				
				//Don't put comma for last field
				if (i < fieldNames.length - 1) {
					sb.append(COMMA);
				}
				
			}
		}
		sb.append(CLOSE_CURLY);
	}
	
	// Represent value of field based on its type
	private static void handleFieldType(FieldType<?> fieldType, Object value, StringBuilder sb) {
		if (fieldType instanceof EventType) {
			processEventField((Event) value, sb);
		} else if (fieldType instanceof SequenceFieldType<?>) {
			processSequenceField((SequenceFieldType<?>)fieldType, (List<?>) value, sb);
		} else if (fieldType instanceof DictionaryFieldType<?, ?>) {
			processDictionaryField((DictionaryFieldType<?, ?>)fieldType, (Map<?, ?>) value, sb);
		} else if (isPrimitiveFieldType(fieldType)) {
			processPrimitiveField(fieldType, value, sb);
		} else {
			throw new IllegalArgumentException("Invalid field type: " + fieldType);
		}
		
	}
	
	// generate field name
	private static void appendFieldname(String fieldName, StringBuilder sb) {
		sb.append(QUOTES+fieldName+QUOTES+":");
	}
	
	// check if field is of primitive type
	private static boolean isPrimitiveFieldType(FieldType<?> fieldType) {
		if (fieldType instanceof BooleanFieldType 
				|| fieldType instanceof IntegerFieldType
				|| fieldType instanceof FloatFieldType 
				|| fieldType instanceof DecimalFieldType
				|| fieldType instanceof StringFieldType) {
			return true;
		}
		return false;
	}
	
	// represent primitive values
	private static void processPrimitiveField(FieldType<?> fieldType, Object value, StringBuilder sb) {
		if (fieldType instanceof StringFieldType) {
			sb.append(QUOTES).append(value).append(QUOTES);
		} else {
			sb.append(value);
		}
	}
	
	// represent value of sequence type fields
	private static void processSequenceField(SequenceFieldType<?> sequenceFieldType, List<?> list, StringBuilder sb) {
		FieldType<?> elementType = sequenceFieldType.getElementType();
		sb.append(OPEN_SQUARE);
		for (int i=0; i < list.size(); i++) {
			handleFieldType(elementType, list.get(i), sb);
			// Don't put comma for last element
			if (i < list.size() - 1) {
				sb.append(COMMA);
			}
		}
		sb.append(CLOSE_SQUARE);
	}
	
	// represent value of sequence type fields
	private static void processDictionaryField(DictionaryFieldType<?,?> dictFieldType, Map<?,?> map, StringBuilder sb) {
		sb.append(OPEN_CURLY);
		FieldType<?> keyType = dictFieldType.getKeyType();
		FieldType<?> valueType = dictFieldType.getValueType();
		
		// key type other than string type not supported
		if (!(keyType instanceof StringFieldType)) {
			return;
		}
		
		Iterator<?> iterator = map.entrySet().iterator();
		while (iterator.hasNext()) {
			Entry<?, ?> entry = (Entry<?, ?>) iterator.next();
				appendFieldname((String) entry.getKey(), sb);
				handleFieldType(valueType, entry.getValue(), sb);
				// dont't put comma for last entry
				if (iterator.hasNext()) {
					sb.append(COMMA);
				}
		}
		sb.append(CLOSE_CURLY);
	}
}
