/// <summary>
/// ESChannelAsyncRequestResponse.cs
/// 
/// This sample demonstrates the AsyncRequestResponse() method of the
/// IEventServiceChannel class.
/// 
/// A ResponseListener will be registered to the IEventServiceChannel and the event
/// AsyncRequestEvent(0, "Hi There") will be sent to the correlator by calling the
/// IEventServiceChannel's AsyncRequestResponse() method and the AsyncResponseEvent() 
/// is expected as a response.  
/// 
/// Requirement:
/// 
/// The RequestResponseEvent.mon must be injected to the correlator before
/// running this example.  Upon receiving the AsyncRequestEvent() event, the
/// AsyncRequestResponseEvent.mon monitorscript will send the response
/// (AsyncResponseEvent()) back to the same channel and hence the callback of the 
/// ResponseListener will be called.
/// 
/// $Copyright(c) 2010, 2012 Progress Software Corporation. All Rights Reserved.$ 
/// $Copyright (c) 2013, 2015 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$ 
/// Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG 
/// 
/// </summary>

using System;

using Apama;
using Apama.Event;
using Apama.Event.Parser;
using Apama.Services.Event;
using Apama.Util;
using System.Threading;

public class ESChannelAsyncRequestResponse {
	private static readonly Logger logger = Logger.GetLogger(typeof(ESChannelAsyncRequestResponse));
	
	// Define the AsyncRequestEvent and AsyncResponseEvent event type
	// NOTE: in order to use the AsyncRequestResponse() method, the Request
	//       and Response event must have a member "messageId" 
	//      (EventServiceChannelConstants.DefaultMessageIdFieldName) of 
	//       MonitorScript type integer, .NET type Int64.
	static readonly Field<long> messageIdField = FieldTypes.Integer.NewField(EventServiceChannelConstants.DefaultMessageIdFieldName);
	static readonly Field<string> msgField = FieldTypes.String.NewField("msg");
	static readonly Field<string> responseMsgField = FieldTypes.String.NewField("responseMsg");
	
	static readonly EventType asyncRequestEventType = new EventType("AsyncRequestEvent", messageIdField, msgField);
	static readonly EventType asyncResponseEventType = new EventType("AsyncResponseEvent", messageIdField, responseMsgField);
	
	private static volatile bool done = false;
	
	public void RunSample(string host, int port) {
		try {

			// get a IEventService instance from the EventServiceFactory
			using (IEventService eventService = EventServiceFactory.CreateEventService(host, port, "ESChannelAsyncRequestResponse sample"))
			{
				// Create the channel on "eventService.sample.channel"
				IEventServiceChannel ourChannel = eventService.AddChannel("eventService.sample.channel", null);
				
				// Register the RequestEventType to the channel
				ourChannel.RegisterEventType(asyncRequestEventType);
				
				// Create the requestEvent
				Event requestEvent = new Event("AsyncRequestEvent(0, \"Hi There\")");
				
				// Now, make the AsyncRequestReponse.  MyReponseListener's 
				// HandleResponse() is the callback method when a response is available
				ourChannel.AsyncRequestResponse(new MyResponseListener(),
												requestEvent,
												asyncResponseEventType);
				
				Console.WriteLine("AsyncRequestEvent() sent...");
				
				// wait to give a chance for response to be received
				while (!done)
					Thread.CurrentThread.Join(500);
			}
			
		} catch (Exception ee) {
			logger.Error("Exception caught: ", ee);
		}
	}
	
	/// <summary> 
	/// Inner class to implement our callback method that is notified when events are recieved.
	/// </summary>
	private class MyResponseListener : IResponseListener {

		/// <summary> Callback method that is called when a matching response Event is 
		/// received for an asynchronous request-response call made on an EventServiceChannel.
		/// 
		/// 
		/// <param name="requestEvent">the original request Event.</param> 
		/// <param name="responseEvent">the corresponding response Event.</param>
		/// </summary>
		public void HandleResponse(Event requestEvent, Event responseEvent) {
			Console.WriteLine("MyResponseListener.HandleResponse() called - ");
			Console.WriteLine("\t requestEvent = " + requestEvent);
			Console.WriteLine("\t responseEvent = " + responseEvent);
			Console.WriteLine("\t response message = " + responseEvent.GetField(responseMsgField));
			
			done = true;
		}
		
		
		/// <summary> Callback method that is called when an exception is thrown within an
		/// asynchronous request-response call made on an EventServiceChannel.
		/// 
		/// <param name="exception">The exception that was thrown in the asynchronous call.</param>
		/// </summary> 
		public void HandleException(Exception exception) {
			logger.Error("Exception caught in MyResponseListener: ", exception);
			// set it to true so that program exits in case of failure
			done = true;
		}
	}
	
	public static void Main(string[] args)
	{
		if ((args.Length == 2) && (int.Parse(args[1]) > 0))
		{
			string host = args[0];
			int port = int.Parse(args[1]);
			new ESChannelAsyncRequestResponse().RunSample(host, port);
		}
		else
		{
			Console.WriteLine("Usage: ESChannelAsyncRequestResponse.exe <host> <port>");
		}
	}

}
