/*
 * StockWatch.java
 *
 * This sample java monitor is used to notify users when a stock trade price falls below
 * a given price floor. An instance of the class can be constructed in two different 
 * modes, viz a top-level or factory mode, and a worker mode. The factory is instantiated 
 * using the default no argument constructor used by the jmon framework on application loading. 
 * This creates workers via the single argument constructor every time a Limit event is 
 * received. The workers perform the price floor monitoring, and are removed when a new Limit 
 * request is received with the same user id and stock name.
 *
 * $Copyright(c) 2002 - 2006 Progress Software Corporation. All Rights Reserved.$ 
 * $Copyright (c) 2013,2019 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$ 
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG 
 *
 * $Revision: 359265 $
 *
 */

import com.apama.jmon.*;


public class StockWatch implements Monitor {
	private static final long serialVersionUID = 25L;
    // the limit event used to instantiate a StockWatch worker 
    private Limit limit;

    // the event expressions and match listeners used in the class instantiation
    private EventExpression dieExpr = null;
    private EventExpression fallExpr = null;



    /**
     * No argument constructor used by the jmon framework on application loading. 
     * Objects instantiated with this constructor become top-level or factory monitors, 
     * looking for limit events to denote a request to start monitoring a stock for a 
     * price floor crosssing
     */
    public StockWatch() {}


    
    /**
     * Construct an instantiation of the class to start monitoring a stock for a 
     * price floor crossing. The constructor uses the Limit event to determine the id 
     * of the user interested in the price fall, the stock name to monitor, and the value 
     * of the price floor. The constructor sets up event expressions and match listeners as 
     * anonymous inner classes to remove the object on receipt of another Limit event by the 
     * same user for the same stock, and to notify the user when the price floor is crossed
     *
     * @param limit  The Limit event used to start monitoring a stock
     */
    public StockWatch(Limit limit) {
        this.limit = limit;
	
	// set up event expression and match listener to effectively remove
	// this object when another Limit event is seen for the same user id 
	// and stock name
	dieExpr= new EventExpression("com.apama.samples.stockwatch.Limit(" + limit.userID + ", \"" + limit.stockName + "\", *)");
        dieExpr.addMatchListener(new MatchListener () {
		private static final long serialVersionUID = 32L;
		public void match(MatchEvent event) {
		    die();
		}
	    });

	// set up event expression and match listener to monitor for the 
	// price floor crossing
	fallExpr= new EventExpression("com.apama.samples.stockwatch.Tick(\"" + limit.stockName + "\", <" + limit.priceFloor + "):tick");
	fallExpr.addMatchListener(new MatchListener () {
		private static final long serialVersionUID = 31L;
		public void match(MatchEvent event) {
		    Tick tick = (Tick)event.getMatchingEvents().get("tick");
		    Match match = new Match(StockWatch.this.limit.userID, StockWatch.this.limit.stockName, tick.price);
		    match.emit();
		    die();
		}
	    });
    }



    /**
     * Implementation of the monitor interface onLoad method. Sets up a top-level 
     * event expression to create new StockWatch instantiations to monitor for 
     * price floor crossings
     */
    public void onLoad() {
	EventExpression factory = new EventExpression("all com.apama.samples.stockwatch.Limit(*,*,>0.0):limit");
	factory.addMatchListener(new MatchListener() {
		private static final long serialVersionUID = 30L;
		public void match(MatchEvent event) {
		    new StockWatch( (Limit)event.getMatchingEvents().get("limit") );
		}
	    });
    } 



    // Helper method to mark this object for garbage collection. Removes all match listeners
    // added to the event expressions, causing the correlator to drop its references to them. 
    // Once done, there is no path to the stockwatch object (previously root -> correlator -> 
    // match listeners -> stockwatch), meaning it can be marked for garbage collection
    private void die(){
	if (dieExpr != null) dieExpr.removeAllMatchListeners();
	if (fallExpr != null) fallExpr.removeAllMatchListeners();
    }
    
}
